<?php

namespace Drupal\entity_merge\Plugin\FieldMergeHandler;

use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\entity_merge\FieldDisplayContext;
use Drupal\entity_merge\FieldMergeContext;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Handles address fields.
 *
 * @FieldMergeHandler(
 *   id = "merge_field_handler_simplenews_subscription",
 *   handles_field_type = "simplenews_subscription",
 *   weight = 100
 * )
 *
 * @package Drupal\entity_merge\Plugin\FieldMergeHandler
 */
class SubscriptionFieldMergeHandler extends DefaultFieldMergeHandler implements ContainerFactoryPluginInterface {

  /**
   * The time service.
   *
   * @var \Drupal\Component\Datetime\TimeInterface
   */
  protected $time;

  /**
   * Construct the subscription field merge handler.
   */
  public function __construct(TimeInterface $time) {
    $this->time = $time;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $container->get('datetime.time')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function displayField(FieldDisplayContext $context): bool {
    /** @var \Drupal\simplenews\Plugin\Field\FieldType\SubscriptionItem[] $secondary_subs */
    $secondary_subs = [];
    foreach ($context->getSecondaryField() as $item) {
      $secondary_subs[$item->target_id] = $item;
    }

    /** @var \Drupal\simplenews\Plugin\Field\FieldType\SubscriptionItem $item */
    foreach ($context->getPrimaryField() as $item) {
      $definition = [
        'comparison' => TRUE,
        'label' => $item->entity->label(),
        'primary' => $item->status ? new TranslatableMarkup('Active') : new TranslatableMarkup('Inactive'),
        'secondary' => '',
      ];
      if (isset($secondary_subs[$item->target_id])) {
        $definition['comparison'] = $secondary_subs[$item->target_id]->status == $item->status || $secondary_subs[$item->target_id]->timestamp > $item->timestamp;
        $definition['secondary'] = $secondary_subs[$item->target_id]->status ? new TranslatableMarkup('Active') : new TranslatableMarkup('Inactive');
        unset($secondary_subs[$item->target_id]);
      }
      $context->setFieldDisplayInfo('', "{$context->getFieldId()}:{$item->target_id}", $definition);
    }

    foreach ($secondary_subs as $item) {
      $context->setFieldDisplayInfo('', "{$context->getFieldId()}:{$item->target_id}", [
        'comparison' => TRUE,
        'label' => $item->entity->label(),
        'primary' => '',
        'secondary' => $item->status ? new TranslatableMarkup('Active') : new TranslatableMarkup('Inactive'),
      ]);
    }

    return self::HANDLED;
  }

  /**
   * {@inheritdoc}
   */
  public function performMerge(FieldMergeContext $context) {
    $primary = $context->getPrimaryField();
    /** @var \Drupal\simplenews\Plugin\Field\FieldType\SubscriptionItem[] $primary_subs */
    $primary_subs = [];
    foreach ($primary as $item) {
      $primary_subs[$item->target_id] = $item;
    }

    /** @var \Drupal\simplenews\Plugin\Field\FieldType\SubscriptionItem $item */
    foreach ($context->getSecondaryField() as $item) {
      // New items simply get added.
      if (!isset($primary_subs[$item->target_id])) {
        $value = $item->getValue();
        if (strpos($value['source'], 'merged') === FALSE) {
          $value['source'] .= ' (merged)';
        }
        $primary->appendItem($value);
      }
      // Merging wants to take the most recent source/timestamp.
      elseif ($item->timestamp > $primary_subs[$item->target_id]->timestamp) {
        $value = $item->getValue();
        if (strpos($value['source'], 'merged') === FALSE) {
          $value['source'] .= ' (merged)';
        }
        $primary_subs[$item->target_id]->setValue($value);
      }

      // Unsubscribe the secondary.
      if ($item->status != SIMPLENEWS_SUBSCRIPTION_STATUS_UNSUBSCRIBED) {
        $item->status = SIMPLENEWS_SUBSCRIPTION_STATUS_UNSUBSCRIBED;
        $item->source = 'merged';
        $item->timestamp = $this->time->getCurrentTime();
      }
    }

    return self::HANDLED;
  }

}
