<?php

namespace Drupal\Tests\entity_merge\Kernel;

use Drupal\Component\Datetime\TimeInterface;
use Drupal\entity_merge\Entity\MergeRequest;
use Drupal\entity_merge\Entity\MergeRequestInterface;
use Drupal\KernelTests\KernelTestBase;

/**
 * Test coverage for the Simplenews subscription field merge handler.
 *
 * @group legacy
 *
 * @coversDefaultClass \Drupal\entity_merge\Plugin\FieldMergeHandler\SubscriptionFieldMergeHandler
 */
class SimplenewsSubscriptionFieldMergeHandlerTest extends KernelTestBase {

  /**
   * Modules to enable.
   *
   * @var array
   */
  public static $modules = [
    'entity_merge',
    'options',
    'user',
    'node',
    'field',
    'system',
    'simplenews',
  ];

  /**
   * {@inheritdoc}
   */
  static protected $configSchemaCheckerExclusions = [
    // Lots of simplenews config fails schema checks.
    'field.storage.simplenews_subscriber.subscriptions',
    'field.field.simplenews_subscriber.simplenews_subscriber.subscriptions',
    'views.view.simplenews_subscribers',
  ];

  /**
   * A newsletter to test with.
   *
   * @var \Drupal\simplenews\NewsletterInterface
   */
  protected $newsletter1;

  /**
   * A second newsletter to test with.
   *
   * @var \Drupal\simplenews\NewsletterInterface
   */
  protected $newsletter2;

  /**
   * The subscriber entity storage.
   *
   * @var \Drupal\Core\Entity\ContentEntityStorageInterface
   */
  protected $subscriberStorage;

  /**
   * {@inheritdoc}
   */
  protected function setUp() {
    parent::setUp();

    $this->installEntitySchema('user');
    $this->installEntitySchema('node');
    $this->installEntitySchema('entity_merge_request');
    $this->installEntitySchema('simplenews_subscriber');

    $this->installSchema('system', 'sequences');
    $this->installSchema('simplenews', 'simplenews_mail_spool');
    $this->installConfig('simplenews');

    $entity_type_manager = $this->container->get('entity_type.manager');
    $this->subscriberStorage = $entity_type_manager->getStorage('simplenews_subscriber');

    $newsletter_storage = $entity_type_manager->getStorage('simplenews_newsletter');
    $this->newsletter1 = $newsletter_storage->create(['id' => 'newsletter_1']);
    $this->newsletter1->save();
    $this->newsletter2 = $newsletter_storage->create(['id' => 'newsletter_2']);
    $this->newsletter2->save();

    $time = $this->prophesize(TimeInterface::class);
    $time->getRequestTime()->willReturn(150);
    $time->getCurrentTime()->willReturn(160);
    $this->container->set('datetime.time', $time->reveal());
  }

  /**
   * Test merging the subscription field data.
   *
   * @param array $subs_1
   *   User 1's initial subscriptions.
   * @param array $subs_2
   *   User 2's initial subscriptions.
   * @param array $expected_subs_1
   *   User 1's final subscriptions.
   * @param array $expected_subs_2
   *   User 2's final subscriptions.
   *
   * @dataProvider dataMerge
   * @covers ::performMerge
   */
  public function testMerge(array $subs_1, array $subs_2, array $expected_subs_1, array $expected_subs_2) {
    // Create the subscriber 1 and it's subscriptions.
    $subscriber_1 = $this->subscriberStorage->create();
    foreach ($subs_1 as $sub) {
      $subscriber_1->get('subscriptions')->appendItem($sub);
    }
    $subscriber_1->save();

    // Create the subscriber 2 and it's subscriptions.
    $subscriber_2 = $this->subscriberStorage->create();
    foreach ($subs_2 as $sub) {
      $subscriber_2->get('subscriptions')->appendItem($sub);
    }
    $subscriber_2->save();

    // Run the merge.
    $request = MergeRequest::create([
      'primary_id' => $subscriber_1->id(),
      'secondary_id' => $subscriber_2->id(),
      'target_type' => 'simplenews_subscriber',
      'status' => MergeRequestInterface::STATUS_NOT_STARTED,
    ]);
    $result = $this->container->get('entity_merge.merger')->run($request);
    $this->assertTrue($result, 'Merge succeeded');

    // Check that we have the correct subscribers for subscriber 1.
    $subscriber_1 = $this->subscriberStorage->loadUnchanged($subscriber_1->id());
    $subscriptions = $subscriber_1->get('subscriptions')->getValue();
    $this->assertEquals($expected_subs_1, $subscriptions, 'Subscriber 1 subscriptions');

    // Check that we have the correct subscribers for subscriber 2.
    $subscriber_2 = $this->subscriberStorage->loadUnchanged($subscriber_2->id());
    $subscriptions = $subscriber_2->get('subscriptions')->getValue();
    $this->assertEquals($expected_subs_2, $subscriptions, 'Subscriber 2 subscriptions');
  }

  /**
   * Data provider for testing subscriber merges.
   */
  public function dataMerge() {
    // Only the subscriber has any subscriptions.
    yield 'only-primary' => [
      'subs_1' => [
        [
          'target_id' => 'newsletter_1',
          'status' => '1',
          'source' => 'source_1',
          'timestamp' => '50',
        ],
      ],
      'subs_2' => [],
      'expected_subs_1' => [
        [
          'target_id' => 'newsletter_1',
          'status' => '1',
          'source' => 'source_1',
          'timestamp' => '50',
        ],
      ],
      'expected_subs_2' => [],
    ];

    // Only the secondary subscriber has any subscriptions.
    yield 'only-secondary' => [
      'subs_1' => [],
      'subs_2' => [
        [
          'target_id' => 'newsletter_1',
          'status' => '1',
          'source' => 'source_2',
          'timestamp' => '50',
        ],
      ],
      'expected_subs_1' => [
        [
          'target_id' => 'newsletter_1',
          'status' => '1',
          'source' => 'source_2 (merged)',
          'timestamp' => '50',
        ],
      ],
      'expected_subs_2' => [
        [
          'target_id' => 'newsletter_1',
          'status' => '0',
          'source' => 'merged',
          'timestamp' => '160',
        ],
      ],
    ];

    // Both subscribers have a single identical subscription.
    yield 'single-identical' => [
      'subs_1' => [
        [
          'target_id' => 'newsletter_1',
          'status' => '1',
          'source' => 'source_1',
          'timestamp' => '50',
        ],
      ],
      'subs_2' => [
        [
          'target_id' => 'newsletter_1',
          'status' => '1',
          'source' => 'source_2',
          'timestamp' => '50',
        ],
      ],
      'expected_subs_1' => [
        [
          'target_id' => 'newsletter_1',
          'status' => '1',
          'source' => 'source_1',
          'timestamp' => '50',
        ],
      ],
      'expected_subs_2' => [
        [
          'target_id' => 'newsletter_1',
          'status' => '0',
          'source' => 'merged',
          'timestamp' => '160',
        ],
      ],
    ];

    // Subscriber 1 has a newer subscription for the single matching newsletter.
    yield 'single-sub-1-newer' => [
      'subs_1' => [
        [
          'target_id' => 'newsletter_1',
          'status' => '1',
          'source' => 'source_1',
          'timestamp' => '60',
        ],
      ],
      'subs_2' => [
        [
          'target_id' => 'newsletter_1',
          'status' => '1',
          'source' => 'source_2',
          'timestamp' => '50',
        ],
      ],
      'expected_subs_1' => [
        [
          'target_id' => 'newsletter_1',
          'status' => '1',
          'source' => 'source_1',
          'timestamp' => '60',
        ],
      ],
      'expected_subs_2' => [
        [
          'target_id' => 'newsletter_1',
          'status' => '0',
          'source' => 'merged',
          'timestamp' => '160',
        ],
      ],
    ];

    // Subscriber 2 has a newer subscription for the single matching newsletter.
    yield 'single-sub-2-newer' => [
      'subs_1' => [
        [
          'target_id' => 'newsletter_1',
          'status' => '1',
          'source' => 'source_1',
          'timestamp' => '50',
        ],
      ],
      'subs_2' => [
        [
          'target_id' => 'newsletter_1',
          'status' => '1',
          'source' => 'source_2',
          'timestamp' => '60',
        ],
      ],
      'expected_subs_1' => [
        [
          'target_id' => 'newsletter_1',
          'status' => '1',
          'source' => 'source_2 (merged)',
          'timestamp' => '60',
        ],
      ],
      'expected_subs_2' => [
        [
          'target_id' => 'newsletter_1',
          'status' => '0',
          'source' => 'merged',
          'timestamp' => '160',
        ],
      ],
    ];

    // Subscriber 2 has no newer subscriptions than subscriber 1.
    yield 'multiple-same-older' => [
      'subs_1' => [
        [
          'target_id' => 'newsletter_1',
          'status' => '1',
          'source' => 'source_1',
          'timestamp' => '50',
        ],
        [
          'target_id' => 'newsletter_2',
          'status' => '1',
          'source' => 'source_1',
          'timestamp' => '50',
        ],
      ],
      'subs_2' => [
        [
          'target_id' => 'newsletter_1',
          'status' => '1',
          'source' => 'source_2',
          'timestamp' => '40',
        ],
      ],
      'expected_subs_1' => [
        [
          'target_id' => 'newsletter_1',
          'status' => '1',
          'source' => 'source_1',
          'timestamp' => '50',
        ],
        [
          'target_id' => 'newsletter_2',
          'status' => '1',
          'source' => 'source_1',
          'timestamp' => '50',
        ],
      ],
      'expected_subs_2' => [
        [
          'target_id' => 'newsletter_1',
          'status' => '0',
          'source' => 'merged',
          'timestamp' => '160',
        ],
      ],
    ];

    // Subscriber 2 has a newer subscriptions than subscriber 1.
    yield 'multiple-same-newer' => [
      'subs_1' => [
        [
          'target_id' => 'newsletter_1',
          'status' => '1',
          'source' => 'source_1',
          'timestamp' => '50',
        ],
        [
          'target_id' => 'newsletter_2',
          'status' => '1',
          'source' => 'source_1',
          'timestamp' => '50',
        ],
      ],
      'subs_2' => [
        [
          'target_id' => 'newsletter_1',
          'status' => '1',
          'source' => 'source_2',
          'timestamp' => '60',
        ],
      ],
      'expected_subs_1' => [
        [
          'target_id' => 'newsletter_1',
          'status' => '1',
          'source' => 'source_2 (merged)',
          'timestamp' => '60',
        ],
        [
          'target_id' => 'newsletter_2',
          'status' => '1',
          'source' => 'source_1',
          'timestamp' => '50',
        ],
      ],
      'expected_subs_2' => [
        [
          'target_id' => 'newsletter_1',
          'status' => '0',
          'source' => 'merged',
          'timestamp' => '160',
        ],
      ],
    ];

    // Subscriber 2 has some subscriptions subscriber 1 does not.
    yield 'multiple-differing' => [
      'subs_1' => [
        [
          'target_id' => 'newsletter_1',
          'status' => '1',
          'source' => 'source_1',
          'timestamp' => '50',
        ],
      ],
      'subs_2' => [
        [
          'target_id' => 'newsletter_1',
          'status' => '1',

          'source' => 'source_2',
          'timestamp' => '40',
        ],
        [
          'target_id' => 'newsletter_2',
          'status' => '1',
          'source' => 'source_2',
          'timestamp' => '60',
        ],
      ],
      'expected_subs_1' => [
        [
          'target_id' => 'newsletter_1',
          'status' => '1',
          'source' => 'source_1',
          'timestamp' => '50',
        ],
        [
          'target_id' => 'newsletter_2',
          'status' => '1',
          'source' => 'source_2 (merged)',
          'timestamp' => '60',
        ],
      ],
      'expected_subs_2' => [
        [
          'target_id' => 'newsletter_1',
          'status' => '0',
          'source' => 'merged',
          'timestamp' => '160',
        ],
        [
          'target_id' => 'newsletter_2',
          'status' => '0',
          'source' => 'merged',
          'timestamp' => '160',
        ],
      ],
    ];
  }

}
